from REFPROPConnector import ThermodynamicPoint, AbstractThermodynamicPoint, RefPropHandler
import unittest


class TestREFPROPConnector(unittest.TestCase):

    def test_init(self):

        ThermodynamicPoint(["air"], ["1"])
        self.assertEqual(True, True)

    def test_list_properties(self):

        tp = ThermodynamicPoint(["air"], ["1"])
        tp.list_properties()
        self.assertEqual(True, True)

    def test_list_units(self):

        tp = ThermodynamicPoint(["air"], ["1"], unit_system="MASS BASE SI")
        tp.list_unit_systems()
        self.assertEqual(True, True)


if __name__ == '__main__':
    unittest.main()
